import re
from Noema._pyExplorer import PyExplorer

class PatternBuilder:
    
    _instance = None  # Variable de classe pour stocker l'instance unique
    _initialized = False  # Indicateur pour contrôler l'initialisation
    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(PatternBuilder, cls).__new__(cls)
        return cls._instance

    def __init__(self, value=None):
        if not self.__class__._initialized:
            self.objects_by_type = {}
            self.noesis_pattern =  self.build_noesis_pattern()
            self.type_white_list = None
            self.type_white_list_names = None
            self.__class__._initialized = True  

    @classmethod
    def instance(cls, value=None):
        if cls._instance is None:
            cls._instance = cls(value)
        return cls._instance
        
    def build_noesis_pattern(self):
        atomicTypesModuleName = "Noema._AtomicTypes"
        names,obj_dict = PyExplorer.extractTypeWhiteList(atomicTypesModuleName)
        self.objects_by_type = obj_dict
        self.type_white_list_names = [name.replace("[", "\[").replace("]", "\]") for name in names]
        self.type_white_list = "|".join(self.type_white_list_names)
        return r"([a-zA-Z_][a-zA-Z0-9_]*)\s*:\s*({})\s*=\s*(f?)\s*\(*\s*((?:'''|\"\"\"|'|\"))\s*(.*?)\s*\4\s*\)*\s*(?:@(\w+)\s*(?:\(\s*(.*?)\s*\))?)?".format(self.type_white_list)

    def whiteListNames(self):
        atomicTypesModuleName = "Noema._AtomicTypes"
        names,obj_dict = PyExplorer.extractTypeWhiteList(atomicTypesModuleName)
        self.type_white_list_names = [name.replace("[", "\[").replace("]", "\]") for name in names]
        return self.type_white_list_names





import importlib
import re
import inspect
import sys
import linecache
import os
from Noema._Generator import Generator
from Noema._noesisModel import NoesisModel
from Noema._patternBuilder import PatternBuilder
from Noema._pyExplorer import PyExplorer
from Noema._pyNoesisBuilder import PyNoesisBuilder
from Noema._noema import Noema

class Noesis:

    def __init__(self):
        self.pattern = PatternBuilder.instance().noesis_pattern
        self.ref_code_lines = []
        self.updated_code_lines = []
        self.ref_model_by_line = {}
        self.updated_model_by_line = {}
        self.marks = []
        self.builder = PyNoesisBuilder()
        self.noema = Noema()
        self.needs_var_update = True
        self.execution_flow = []
    
    def constitute(self, subject, verbose=False):
        self.verbose = verbose
        self.subject = subject
        self.previous_state = str(self.subject.llm)
        self.subject.llm.reset()
        infos = self.builder.build_noesis(self.description,sender=self)
        self.subject.parents.append(self)
        self.subject.llm += infos["noesis"]
        self.ref_code_lines = infos["code_ref"]
        self.ref_model_by_line = infos["model_ref"]
        self.updated_code_lines = infos["code_updated"]
        self.updated_model_by_line = infos["model_updated"]
        return self.run()
        
    def produceValue(self, model, local_vars) -> dict:
        return self.noema.generateFromModel(model,self.subject, local_vars=local_vars, verbose=self.verbose)
        
    def append_flow(self, lineno, model, produced_value):
        self.execution_flow.append((lineno, model, produced_value))
        
    def trace_func(self, frame, event, arg):
        if frame.f_code.co_name != "updated_description":
            return None
        if event == 'line':
            lineno = frame.f_lineno
            line = linecache.getline(frame.f_code.co_filename, lineno)
            try:
                if lineno in self.updated_model_by_line:
                    local_vars = frame.f_locals
                    if self.needs_var_update:
                        for k,model in self.updated_model_by_line.items():
                            for key, value in model.var_and_values.items():
                                local_vars[model.variable].__dict__[key] = value
                        self.needs_var_update = False
                    if self.updated_model_by_line[lineno].variable in local_vars:
                        model = self.updated_model_by_line[lineno]
                        if isinstance(local_vars.get(model.variable), Generator):
                            produced_value = None
                            original_value = model.original_value
                            if model.type == "Fill":
                                original_value = model.original_value
                                goal = original_value[0]
                                fill = original_value[1]
                                goal = goal.format(**local_vars)
                                white_list = PatternBuilder.instance().whiteListNames()
                                for type_name in white_list:
                                    fill = re.sub(
                                        r"\{" + type_name + r"(:[_a-zA-Z][_a-zA-Z0-9]*)?\}",
                                        r"##" + type_name + r"\1##",
                                        fill,
                                    )
                                    
                                fill = fill.format(**local_vars)
                                model.value = (goal,fill)
                                produced_value = self.produceValue(model,local_vars=local_vars)
                                
                                for key, value in model.var_and_values.items():
                                    local_vars[model.variable].__dict__[key] = value
                            else:
                                original_value = model.original_value
                                original_value = original_value.format(**local_vars)        
                                model.value = str(original_value)
                                produced_value = self.produceValue(model,local_vars=local_vars)
                            
                            local_vars[model.variable].value = produced_value["value"]
                            local_vars[model.variable].noesis = produced_value["noesis"]
                            local_vars[model.variable].noema = produced_value["noema"]
                        self.append_flow(lineno, model, produced_value=produced_value)
                    # else:
                    #     print(f"Var {self.updated_model_by_line[lineno].variable} not found in local vars")
            except Exception as e:
                pass
                # print(f"Error while tracing line {lineno}: {e}")
                # print("Line", line)
        return self.trace_func

    def run(self):
        def local_trace(frame, event, arg):
            return self.trace_func(frame, event, arg)        
        namespace = PyExplorer.create_name_space(self)
        updated_code = "\n".join(self.updated_code_lines)
        method_name = "updated_description"
        exec(updated_code, globals(), namespace)
        updated_description = namespace[method_name]
        setattr(self, method_name, updated_description)
        globals().update(namespace)
        try:
            sys.settrace(local_trace)
            exec(f"result = self.{method_name}(self)", globals(), namespace)
            result = namespace.get('result')
            self.subject.parents.pop()
            return result
        except Exception as e:
            print(f"Error while executing '{method_name}': {e}")
        finally:
            if len(self.subject.parents) == 1:
                self.subject.llm.reset()
                self.subject.llm += self.previous_state
                sys.settrace(None)
                
    # def build_graph(self):
    #     graph = "graph TD\n"
    #     for i, (lineno, model, produced_value) in enumerate(self.execution_flow):
    #         variable = model.variable
    #         value = produced_value['value']
    #         graph += f'    {i}["{variable} = {value}"]\n'
    #         if i > 0:
    #             prev_lineno, prev_model, prev_produced_value = self.execution_flow[i - 1]
    #             graph += f'    {i-1} --> {i}\n'
    #     return graph


from ._AtomicGenerator import AtomicGenerator
from ._noesis import *
from ._AtomicTypes import *
from ._pyExplorer import PyExplorer
from ._patternBuilder import PatternBuilder
from ._noema import Noema
from ._CustomGenerator import CustomGenerator
from ._Subject import Subject


class NoesisModel:
    
    def __init__(self,variable,type,value,annotation=None,var_and_values={}):
        self.variable = variable
        self.type = type
        self.original_value = value
        self.value = value
        self.annotation = annotation
        self.next_model = None
        self.var_and_values = var_and_values
        
    def display_var(self):
        return "#"+self.variable.upper()+":"
    
    def stops(self):
        if self.next_model is not None:
            v = self.next_model.variable
            return [self.next_model.display_var()]
        else:
            return []
    
    def __repr__(self):
        return f"{self.variable} {self.type} = {self.value} @{self.annotation} | {self.var_and_values}"
    



from guidance import models

from Noema._noema import Noema

class Subject:
    def __init__(self, llm_path):
        self.data = {}
        self.namespace_stack = []
        self.parents = []
        self.llm = models.LlamaCpp(
            llm_path,
            n_gpu_layers=99,
            n_ctx=512*3,
            echo=False,
        )
        self.noesis = ""
        self.noema = Noema()
            
    def __str__(self):
        return str(self.llm)


import re
from typing import Annotated
from annotated_types import MinLen
from guidance import capture, gen
from pydantic import TypeAdapter
from guidance import json as jj
from Noema._AtomicGenerator import AtomicGenerator
from Noema._CustomGenerator import CustomGenerator
from Noema._patternBuilder import PatternBuilder
from Noema.cfg import *
from guidance import models, gen, select, capture, substring
from Noema._Generator import Generator
from Noema._patternBuilder import PatternBuilder
from Noema.cfg import G
    
class Email(CustomGenerator):
    grammar = G.email()
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.email(), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)

    
class Url(AtomicGenerator):
    grammar = G.url_http_https()
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.url_http_https(), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)


class Information(AtomicGenerator):
    grammar = None
    return_type = str

class Instruction(AtomicGenerator):
    grammar = None
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        state.llm += "[INST]" + noesis_model.value + "[/INST]\n"
        return "[INST]" + noesis_model.value + "[/INST]\n"


class Paragraph(AtomicGenerator):
    grammar = G.alphaNumPunctForParagraph()
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += "-->"+noesis_model.value + "\n"
        if noesis_model.next_model is not None:
            llm += noesis_model.display_var() + " " + gen(name="response",stop=noesis_model.stops())
        else:
            llm += noesis_model.display_var() + " " + gen(name="response",stop=["\n"])
        res = llm["response"]
        res = res.strip()    
        res = res.replace(noesis_model.display_var(),"")        

        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)

class Fill(AtomicGenerator):
    grammar = G.alphaNumPunctForParagraph()
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += "\n"+noesis_model.value[0] + "\n"
        # replacing ##Type:Name## with <blank_Type> in noesis_model.value[1] extracting the type put in the blank
        prefill = re.sub(r"##(\w+):(\w+)##", r"<blank>", noesis_model.value[1])
        llm += "#Fill-in-the-blank:\n"+prefill+"\nI will rewrite the following, replacing the <blank> values:\n " "\n"
        value = noesis_model.value[1]
        white_list_names = PatternBuilder.instance().whiteListNames()
        pattern = r"(?P<static>[^\#]+)|##(?P<type>\w+)(?::(?P<name>\w+))?##"
        matches = re.finditer(pattern, value)
        value = value.format(**local_vars)
        elements = {}
        full_response = ""
        for match in matches:
            static_part = match.group("static")
            type_name = match.group("type")
            var_name = match.group("name")

            if static_part:
                u_static_part = static_part
                llm += u_static_part
                full_response += u_static_part

            if type_name:
                if type_name in globals():
                    instance = globals()[type_name]()
                    res = None
                    if type(instance) is IntList:
                        llm += capture(G.arrayOf(G.positive_num()), name="response")
                        res = llm["response"]
                        if var_name:
                            _res = res[1:-1].split(",")
                            _res = [int(el.strip()[1:-1]) for el in _res]
                            _res = [el for el in _res if el]
                            elements[var_name] = _res
                    elif type(instance) is SentenceList:
                        ta = TypeAdapter(Annotated[list[str], MinLen(10)])
                        llm +=  jj(name="response", schema=ta)
                        res = llm["response"]
                        if var_name:
                            _res = res[1:-1].split(",")
                            _res = [str(el.strip()[1:-1]) for el in _res]
                            _res = [el for el in _res if el]
                            elements[var_name] = _res
                            
                    elif type(instance) is WordList:
                        llm += capture(G.arrayOf(G.word()), name="response")
                        res = llm["response"]
                        if var_name:
                            _res = res[1:-1].split(",")
                            _res = [str(el.strip()[1:-1]) for el in _res]
                            _res = [el for el in _res if el]
                            elements[var_name] = _res
                    elif type(instance) is BoolList:
                        llm += capture(G.arrayOf(G.bool()), name="response")
                        res = llm["response"]
                        if var_name:
                            _res = res[1:-1].split(",")
                            _res = [str(el.strip()[1:-1]).lower() == "yes" for el in _res]
                            _res = [el for el in _res if el]
                            elements[var_name] = _res
                    elif type(instance) is FloatList:
                        llm += capture(G.arrayOf(G.float()), name="response")
                        res = llm["response"]
                        if var_name:
                            _res = res[1:-1].split(",")
                            _res = [float(el.strip()[1:-1]) for el in _res]
                            _res = [el for el in _res if el]
                            elements[var_name] = _res
                    elif type(instance) is SubString:
                        llm += substring(str(state.llm), name="response")
                        res = llm["response"]
                        if var_name:
                            elements[var_name] = res
                    elif issubclass(type(instance), CustomGenerator):   
                        llm += gen(name="response", stop=instance.stops, regex=instance.regex) + "\n"
                        res = llm["response"]
                        if var_name:
                            elements[var_name] = instance.return_type(res)
                    elif issubclass(type(instance), CodeGenerator):
                        llm += "\n Produce only the code, no example or explanation." + "\n"
                        llm += f" ```{self.__class__.__name__}\n" + gen(stop="```",name="response")        
                        res = llm["response"]
                        if var_name:
                            elements[var_name] = res
                    else:
                        llm += capture(instance.grammar, name="response")  
                        res = llm["response"]
                        if var_name:
                            if type_name == "Bool":
                                elements[var_name] = res.lower() == "yes"
                            elif type_name == "Float":
                                elements[var_name] = float(res)
                            elif type_name == "Int":
                                elements[var_name] = int(res)
                            else:
                                elements[var_name] = res
                                
                    full_response += res
                else:
                    print(f"Type {type_name} not found")
        noesis_model.var_and_values = elements        
        state.llm += noesis_model.display_var() + " " + full_response + "\n"
        return full_response
    
    def extract_elements(input_string):
        pattern = r"([^{}]+)|{(\w+)}"
        matches = re.findall(pattern, input_string)

        elements = []
        for static_part, dynamic_part in matches:
            if static_part:
                elements.append(static_part.strip())
            elif dynamic_part:
                elements.append(dynamic_part)

        return elements


class Int(AtomicGenerator):
    grammar = G.num()
    return_type = int
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.num(), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)
    
class Word(AtomicGenerator):
    grammar = G.word()
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        if self.grammar is None:
            state.llm += noesis_model.display_var() + noesis_model.value + "\n"
            return self.return_type(noesis_model.value)
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.word(), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)
    
class Sentence(AtomicGenerator):
    grammar = G.sentence()
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        if self.grammar is None:
            state.llm += noesis_model.display_var() + noesis_model.value + "\n"
            return self.return_type(noesis_model.value)
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.sentence(), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)
    
class Bool(AtomicGenerator):
    grammar = G.bool()
    return_type = bool
    
    def execute(self, noesis_model, state, local_vars = None):
        if self.grammar is None:
            state.llm += noesis_model.display_var() + noesis_model.value + "\n"
            return self.return_type(noesis_model.value)
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.bool(), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)
    
class Float(AtomicGenerator):
    grammar = G.float
    return_type = float

    def execute(self, noesis_model, state, local_vars = None):
        if self.grammar is None:
            state.llm += noesis_model.display_var() + noesis_model.value + "\n"
            return self.return_type(noesis_model.value)
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.float(), name="response")        
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)    

class CodeGenerator(AtomicGenerator):
    grammar = G.alphaNumPunct()
    return_type = str

    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += noesis_model.value + " Produce only the code, no example or explanation." + "\n"
        llm += noesis_model.display_var() + f" ```{self.__class__.__name__}\n" + gen(stop="```",name="response")        
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)
    
class SentenceList(AtomicGenerator):
    grammar = G.sentence()
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += noesis_model.value + "\n"
        ta = TypeAdapter(Annotated[list[str], MinLen(10)])
        llm += noesis_model.display_var() + " " + jj(name="response", schema=ta)
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        res = res[1:-1].split(",")
        res = [self.return_type(el.strip()[1:-1]) for el in res]
        # remove empty strings
        res = [el for el in res if el]
        return res
    
    
class BoolList(AtomicGenerator):
    grammar = G.sentence()
    return_type = bool
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.arrayOf(G.bool()), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        res = res[1:-1].split(",")
        res = [str(el.strip()[1:-1]).lower() == 'yes' for el in res]
        return res
    
class FloatList(AtomicGenerator):
    grammar = G.sentence()
    return_type = float
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.arrayOf(G.float()), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        res = res[1:-1].split(",")
        res = [self.return_type(el.strip()[1:-1]) for el in res]
        return res
    
class WordList(AtomicGenerator):
    grammar = G.sentence()
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.arrayOf(G.word()), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        res = res[1:-1].split(",")
        res = [self.return_type(el.strip()[1:-1]) for el in res]
        res = [el for el in res if el]
        return res
    
class IntList(AtomicGenerator):
    grammar = G.sentence()
    return_type = int
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + capture(G.arrayOf(G.num()), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        res = res[1:-1].split(",")
        res = [self.return_type(el.strip()[1:-1]) for el in res]
        return self.return_type(res)


class Reflexion(Generator):
    
    def __init__(self, ):
        pass
    
    def execute(self, noesis_model, state, local_vars = None):
        noema = ""
        prompt = f"""[INST]{noesis_model.value}\n Follow these steps of reasoning, using a loop to determine whether the process should continue or if the reflection is complete:
1. Initial Hypothesis: Provide a first answer or explanation based on your current knowledge.
2. Critical Analysis: Evaluate the initial hypothesis. Look for contradictions, weaknesses, or areas of uncertainty.
3. Conceptual Revision: Revise or improve the hypothesis based on the critiques from the Critical Analysis.
4. Extended Synthesis: Develop a more complete and nuanced response by incorporating additional perspectives or knowledge.
5. Loop or Conclusion: Return to the Decision Point. If the answer is now coherent and well-justified, you repond 'satisfying' and move to the Conclusion. If further refinement is needed, respond 'loop again' and go to the Critical Analysis. 
6. Final Conclusion: Once the reflection is considered complete, provide a final answer, clearly explaining why this response is coherent and well-justified, summarizing the key steps of the reasoning process.
7. Quality of the reflection: Provide a quality assessment of the reflection process. 
Done.
[/INST]

"""
        lm_copy = str(state.llm)
        lm = state.llm
        lm += prompt + "1. Initial Hypothesis: " + gen(name="reflexion",stop=["2. Critical Analysis:"]) 
        noema += "\n        ***Initial Hypothesis: " + lm["reflexion"].strip()
        counter = 2
        loop_count = 0
        while True:
                lm += f"{counter}. Critical Analysis:" 
                counter += 1
                lm += gen(name="critic",stop=[f"{counter}. Conceptual Revision:"]) 
                noema += "\n"+lm["critic"].strip()
                lm += f"{counter}. Conceptual Revision: " 
                counter += 1
                lm += gen(name="conceptual",stop=[f"{counter}. Extended Synthesis:"]) 
                noema += "\n" + lm["conceptual"].strip()
                lm += f"{counter}. Extended Synthesis: " 
                counter += 1
                lm += gen(name="synthesis",stop=[f"{counter}. Loop or Conclusion:"]) 
                noema += "\n" + lm["synthesis"].strip()
                lm += f"{counter}. Loop or Conclusion: " 
                counter += 1
                lm += capture(select(['satisfying', 'loop again']),name="finished")+ "\n"
                noema += "\n" + lm["finished"].strip()
                if lm["finished"] == "satisfying":
                    lm += f"{counter}. Final Conclusion: " 
                    counter += 1
                    lm += gen(name="response",stop=[f"{counter}. Quality of the reflection:"])
                    noema += "\n" + lm["response"].strip()
                    break
                loop_count += 1
                if loop_count >= 1:
                    lm += f"{counter}. Final Conclusion: " 
                    counter += 1
                    lm += gen(name="response",stop=[f"{counter}. Quality of the reflection:"])
                    noema += "\n" + lm["response"].strip()
                    break
                
        noema += "\nReflexion loop completed."
        res = lm["response"]
        state.llm.reset()
        state.llm += lm_copy+"\n"
        state.llm += noesis_model.display_var() + res + "\n"
        return res,noema, prompt



class Select(AtomicGenerator):
    grammar = G.sentence()
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        # si noesis_model.value n'est pas de type list
        if type(noesis_model.value) is not list:
            raise Exception("Select value must be a list")
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + select(noesis_model.value, name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)

class SubString(AtomicGenerator):
    grammar = G.sentence()
    return_type = str
    
    def execute(self, noesis_model, state, local_vars = None):
        llm = state.llm
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + substring(state.llm, name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)

class Python(CodeGenerator):
    pass

class Java(CodeGenerator):
    pass

class C(CodeGenerator):
    pass

class Cpp(CodeGenerator):
    pass

class CSharp(CodeGenerator):
    pass

class JavaScript(CodeGenerator):
    pass

class TypeScript(CodeGenerator):
    pass

class HTML(CodeGenerator):
    pass

class CSS(CodeGenerator):
    pass

class SQL(CodeGenerator):
    pass

class NoSQL(CodeGenerator):
    pass

class GraphQL(CodeGenerator):
    pass

class Rust(CodeGenerator):
    pass

class Go(CodeGenerator):
    pass

class Ruby(CodeGenerator):
    pass

class PHP(CodeGenerator):
    pass

class Shell(CodeGenerator):
    pass

class Bash(CodeGenerator):
    pass

class PowerShell(CodeGenerator):
    pass

class Perl(CodeGenerator):
    pass

class Lua(CodeGenerator):
    pass

class R(CodeGenerator):
    pass

class Scala(CodeGenerator):
    pass

class Kotlin(CodeGenerator):
    pass

class Dart(CodeGenerator):
    pass

class Swift(CodeGenerator):
    pass

class ObjectiveC(CodeGenerator):
    pass

class Assembly(CodeGenerator):
    pass

class VHDL(CodeGenerator):
    pass

class Verilog(CodeGenerator):
    pass

class SystemVerilog(CodeGenerator):
    pass

class Julia(CodeGenerator):
    pass

class MATLAB(CodeGenerator):
    pass

class COBOL(CodeGenerator):
    pass

class Fortran(CodeGenerator):
    pass

class Ada(CodeGenerator):
    pass

class Pascal(CodeGenerator):
    pass

class Lisp(CodeGenerator):
    pass

class Prolog(CodeGenerator):
    pass

class Smalltalk(CodeGenerator):
    pass

class APL(CodeGenerator):
    pass

class Action:
    pass


from guidance import models, gen, select, capture
from Noema._Generator import Generator

class CustomGenerator(Generator):
    
    regex = None
    return_type = None
    stops = ["\n"]
    
    def execute(self, noesis_model, state):
        if self.regex is None or self.return_type is None:
            raise Exception("regex and return_type must be set in the CustomGenerator class")
        llm = state.llm
        print(noesis_model.value)
        llm += noesis_model.value + "\n"
        llm += noesis_model.display_var() + " " + gen(regex=self.regex,name="response",stop_regex=" ", max_tokens=100) + "\n"
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)



import sys
import os
import unittest

# Add the project directory to the sys.path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from Noema import *

class TestNoema(unittest.TestCase):
    def setUp(self):
        self.subject = Subject("../Models/Mistral-NeMo-Minitron-8B-Instruct.Q4_K_M.gguf")
    
    # Test Horizon creation
    # Test var creation and storage
    # Test Sentence creation
    # Test Int creation
    # Test Float creation
    # Test Bool creation
    # Test Select creation
    # Test IF/ELSE
    # Test Sentence creation from var
    # Test Int creation from var
    # Test Float creation from var
    # Test Bool creation from var
    # Test Select creation from var
    # Test IF/ELSE from var
    # Ajouter un SelectOneOrMore
    
if __name__ == '__main__':
    unittest.main()



from guidance import gen
from Noema._AtomicTypes import *
from Noema._patternBuilder import PatternBuilder
from Noema._pyExplorer import PyExplorer


class Noema:

    def __init__(self):
        process = []
        
    def action_needed(self, model, state, local_vars, verbose):
        lm = state.llm
        lm += f"\nExtract the main action from '{model.value}'\n"
        lm += "Action: " + gen(stop=[".","\n"],name="action") + "\n"
        print("Actions: ", lm["action"])
        functions = state.memoir.retrieves(lm["action"], subject='function')
        if len(functions) > 0:
            selected_func = functions[0]
            func_name = selected_func.function_name()
            parameter_names = selected_func.parameters_names()
            print("Function name: ", func_name)
            print("Parameter names: ", parameter_names)
            
            lm += f"To do '{model.value}', you need to execute the following function: \n" + selected_func.value + "\n"
            lm += "Function call (with respect to the doc string): \n"
            lm += "res = "+func_name + "("

            parameters_values = []
            for i in range(len(parameter_names)):
                pName = parameter_names[i]
                lm += pName + "="
                stops = []
                if i == len(parameter_names) - 1:
                    stops = [")"]
                else:
                    stops = [parameter_names[i+1]+"=", ")"]
                lm += gen(stop=stops,name="p"+str(i))
                pValue = lm["p"+str(i)].strip()
                if pValue[-1] == ",":
                    pValue = pValue[:-1]
                parameters_values.append(pName+"="+pValue)
            print("Parameters values: ", parameters_values)
            print("Call will be: ", func_name + "(" + ",".join(parameters_values) + ")")
            
        state.llm += model.display_var() + "FUNCTION RES" + "\n"
        return { "value": "FUNCTION RES", "noesis": model.value, "noema": model.value, "variables":{} }
    
    def gen_atomic(self, model, subject, local_vars, verbose):
        instance = obj = PatternBuilder.instance().objects_by_type[model.type]()
        if model.type == "Fill":
            res = instance.execute(noesis_model=model,state=subject,local_vars=local_vars)
            if verbose:
                print(f"{model.variable} = \033[93m{res}\033[0m (\033[94m{model.value[0]}\033[0m)")
            return { "value": res, "noesis": model.value, "noema": model.value, "variables":{}}
        elif model.type == "Reflexion":
            res, noema, prompt = Reflexion().execute(noesis_model=model,state=subject)
            if verbose:
                print(f"{model.variable} = \033[93m{res}\033[0m (\033[94m{prompt}\033[0m)")
            return { "value": res, "noesis": model.value, "noema": noema, "variables":{}}
        elif model.type == "Information":
            res = instance.execute(noesis_model=model,state=subject)
            if verbose:
                print(f"{model.variable} = \033[93m{res}\033[0m (\033[94mINFORMATION\033[0m)")
            return { "value": res, "noesis": model.value, "noema": model.value, "variables":{} }
        else:
            res = instance.execute(noesis_model=model,state=subject)
            if verbose:
                print(f"{model.variable} = \033[93m{res}\033[0m (\033[94m{model.value}\033[0m)")
            return { "value": res, "noesis": model.value, "noema": model.value, "variables":{}} 

    def generateFromModel(self, model, subject, local_vars, verbose) -> dict:
        type = model.type
        res = None
        if model.annotation == "Action":
            return self.action_needed(model, subject, local_vars, verbose)
        else:
            return self.gen_atomic(model, subject, local_vars, verbose)


from setuptools import setup, find_packages

# Lire le contenu du README.md
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='Noema',
    version='1.0.4',
    description='Description of Noema',
    long_description=long_description,  # Inclure la description longue
    long_description_content_type='text/markdown',  # Spécifiez le format de la description (markdown ou rst)
    author='Alban Perli',
    author_email='alban.perli@gmail.com',
    url='https://github.com/AlbanPerli/Noema-Declarative-AI',
    packages=find_packages(),
    install_requires=[
        'llama-cpp-python',
        'guidance',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)





import ast
import inspect
from io import StringIO
import re
import tokenize

from Noema._noesisModel import NoesisModel
from Noema._patternBuilder import PatternBuilder


class PyNoesisBuilder:
    
    pattern = PatternBuilder.instance().noesis_pattern
    
    def __init__(self):
        pass
        
    def build_noesis(self, description, sender=None):
        ref_code_lines = []
        ref_model_by_line = {}
#         noesis = """[SYSTEM] You are functioning like an hybrid of an Assistant and a python interpreter.
# For the python interpreter, instructions are provided in the form of a python program.
# For the Assistant, special instruction begins with #TASK_NAME: followed by the instruction that needs to be executed.
# [/SYSTEM]"""
        noesis = ""
        source = inspect.getsource(description)        
        system_prompt = inspect.getdoc(description)
        noesis += "\n<s>[INST]"+system_prompt+"\nYou are functionning in a loop of though. Your response can be multiline. Here is the reasonning you are currently executing:\n\n"
        source = self.clean_code(source)
        code_lines = source.splitlines()
        code_lines = [line for line in code_lines if line.strip()]
        ref_code_str = "\n".join(code_lines)
        ref_code_str,models = self.extract_fill_declarations_ast(ref_code_str)
        ref_model_by_line.update(models)
        code_lines = ref_code_str.splitlines()
        ref_code_lines = code_lines
        marks = []
        context = {"self":sender}
            
        for i in range(len(code_lines)):
            line = code_lines[i]
            current_indent = len(line) - len(line.lstrip())
            indent_discrete = 0
            line = line.strip()
            if i in ref_model_by_line:
                model = ref_model_by_line[i]
                noesis += f"{model.display_var()} {model.value[0]}\n"
            elif re.match(self.pattern, line):
                model = self.extract_and_evaluate(line,context)
                if model.annotation != None:
                    pass
                if model.value[-1] not in ['.', '!', '?']:
                    model.value += '.'
                noesis += f"{model.display_var()} {model.value}\n"
                ref_model_by_line[i] = model
                
            # else:
            #    noesis += " "*current_indent+ f"{line}\n"
        noesis += "[/INST]\n\n"
        updated_code, updated_model = self.update_code(ref_code_lines, ref_model_by_line)
        for key in updated_model:
            if key+1 in updated_model:
                updated_model[key].next_model = updated_model[key+1]
                
        return {"noesis":noesis, "code_ref":ref_code_lines, "model_ref":ref_model_by_line, "code_updated":updated_code, "model_updated":updated_model, "marks":marks}
    
    
    def update_code(self, ref_code_lines, ref_model_by_line):
        updated_model_by_line = {}
        updated_code_lines = ["def updated_description(self):"]
        for key in ref_model_by_line:
            updated_code_lines.append(ref_model_by_line[key].variable+" = Generator()")
        instruction_nb = len(updated_code_lines)
        updated_code_lines.extend(ref_code_lines)
        for key in ref_model_by_line:
            updated_model_by_line[key+instruction_nb+1] = ref_model_by_line[key]
            current_line = updated_code_lines[key+instruction_nb]
            nb_spaces = len(current_line) - len(current_line.lstrip())
            updated_code_lines[key+instruction_nb] = " "*nb_spaces + 'pass'
        updated_code_lines = ["    "+line if i != 0 else line for i, line in enumerate(updated_code_lines)]
        return updated_code_lines, updated_model_by_line

    def extract_fill_declarations_ast(self, code):
        tree = ast.parse(code) 
        results = {}
        lines = code.splitlines()
        updated_lines = lines.copy()

        class FillVisitor(ast.NodeVisitor):
            def visit_AnnAssign(self, node):
                if (
                    isinstance(node.annotation, ast.Name)
                    and node.annotation.id == 'Fill'
                    and isinstance(node.value, ast.Tuple)
                ):
                    variable_name = node.target.id

                    start_line = node.lineno - 1 
                    end_line = node.end_lineno - 1 
                    line_to_remove = updated_lines[start_line:end_line + 1]

                    original_line = lines[start_line]
                    indent = original_line[:len(original_line) - len(original_line.lstrip())]

                    tuple_lines = lines[start_line:end_line + 1]

                    tuple_content = '\n'.join(tuple_lines)
                    open_paren_idx = tuple_content.find('(')
                    close_paren_idx = tuple_content.rfind(')')
                    tuple_content = tuple_content[open_paren_idx + 1 : close_paren_idx]
                    
                    try:
                        values = self.extract_tuple_strings(tuple_content)
                    except ValueError as ve:
                        print(f"Error while extracting tuple strings: {ve}")
                        return

                    fill = values[1]
                    white_list = PatternBuilder.instance().whiteListNames()
                    for type_name in white_list:
                        fill = re.sub(
                            r"\{" + type_name + r"(:[_a-zA-Z][_a-zA-Z0-9]*)?\}",
                            r"##" + type_name + r"\1##",
                            fill,
                        )
                    pattern = r"(?P<static>[^\#]+)|##(?P<type>\w+)(?::(?P<name>\w+))?##"
                    matches = re.finditer(pattern, fill)
                    elements = {}
                    full_response = ""
                    var_and_values = {}
                    for match in matches:
                        static_part = match.group("static")
                        type_name = match.group("type")
                        var_name = match.group("name")
                        if var_name != None:
                            var_and_values[var_name] = None
                        
                    model = NoesisModel(variable_name, "Fill", values, var_and_values=var_and_values)
                    results[start_line] = model
                    
                    for i in range(len(updated_lines)):
                        if updated_lines[i] in line_to_remove:
                            updated_lines[i] = f"{indent}pass"
                    
            def extract_tuple_strings(self, tuple_content):
                tuple_str = f"({tuple_content})"
                tokens = tokenize.generate_tokens(StringIO(tuple_str).readline)
                strings = []
                for toknum, tokval, _, _, _ in tokens:
                    if toknum == tokenize.STRING:
                        if tokval.startswith(('f"', "f'", 'F"', "F'")):
                            tokval = tokval[1:]
                        elif tokval.startswith(('u"', "u'", 'U"', "U'")):
                            tokval = tokval[1:]
                        elif tokval.startswith(('r"', "r'", 'R"', "R'")):
                            tokval = tokval[1:]
                        elif tokval.startswith(('fr"', "fr'", 'FR"', "FR'")):
                            tokval = tokval[2:]
                        elif tokval.startswith(('rf"', "rf'", 'RF"', "RF'")):
                            tokval = tokval[2:]
                        if tokval.startswith('"""') and tokval.endswith('"""'):
                            s = tokval[3:-3]
                        elif tokval.startswith("'''") and tokval.endswith("'''"):
                            s = tokval[3:-3]
                        elif tokval.startswith('"') and tokval.endswith('"'):
                            s = tokval[1:-1]
                        elif tokval.startswith("'") and tokval.endswith("'"):
                            s = tokval[1:-1]
                        else:
                            s = tokval  
                        strings.append(s)
                if len(strings) != 2:
                    raise ValueError("Fill object must contain exactly two strings")
                return tuple(strings)

        FillVisitor().visit(tree)
        updated_lines = [line for line in updated_lines if line.strip()]
        return "\n".join(updated_lines), results
    
    
    

    def extract_and_evaluate(self, text, context):
        results = []        
        matches = re.finditer(self.pattern, text)
        for match in matches:
            var_name = match.group(1)  
            var_type = match.group(2)  
            f_prefix = match.group(3)  
            value = match.group(5)     
            annotation_name = match.group(6)  
            annotation_args = match.group(7)

            if f_prefix == 'f':
                evaluated_value = value
                # try:
                #     evaluated_value = eval(f'f"{value}"', {}, context)
                # except Exception as e:
                #     evaluated_value = f"Eval Error: {e}"
            else:
                evaluated_value = value

            if annotation_name:
                if annotation_args:
                    annotation = f"{annotation_name}({annotation_args})"
                else:
                    annotation = annotation_name
            else:
                annotation = None

            results.append(NoesisModel(var_name, var_type, evaluated_value, annotation))

        return results[0] if results else None


    def clean_code(self, code):
        source = inspect.cleandoc(code)
        source = re.sub(r'"""(.*?)"""', '', source, count=1, flags=re.DOTALL)
        source = re.sub(r'#.*', '', source)
        source = "\n".join([line for line in source.splitlines() if line.strip()])
        source = re.sub(r'def\s+\w+\(.*?\):', '', source)
        return source

    
    
    



from guidance import models, gen, select, capture
from Noema._Generator import Generator

class AtomicGenerator(Generator):
    
    grammar = None
    return_type = None

    def execute(self, noesis_model, state, local_vars = None):
        if self.grammar is None:
            state.llm += noesis_model.display_var() + noesis_model.value + "\n"
            return self.return_type(noesis_model.value)
        llm = state.llm
        llm += noesis_model.display_var() + " " + noesis_model.value + " " + capture(self.grammar(), name="response")
        res = llm["response"]
        state.llm += noesis_model.display_var() + " " + res + "\n"
        return self.return_type(res)



import importlib
import inspect
import re

class PyExplorer:
    
    @staticmethod
    def extractTypeWhiteList(module_name):
        class_names, object_dict = PyExplorer._extract_class_names_and_objects(module_name)
        objects = []
        for key in object_dict:
            objects.append(object_dict[key])
        
        return class_names, object_dict
    
    @staticmethod
    def _extract_class_names_and_objects(module_name):
        module = importlib.import_module(module_name)
        class_names = []
        object_dict = {}
        for name, obj in inspect.getmembers(module, inspect.isclass):
            if obj.__module__ == module_name:
                class_names.append(name)
                object_dict[name] = obj
        return class_names, object_dict
    
    
    @staticmethod
    def import_from_string(import_statement, namespace) -> dict :
        if import_statement.startswith("from "):
            parts = import_statement.split()
            module_name = parts[1]
            attribute_name = parts[3]
            if attribute_name == "*":
                namespace = PyExplorer.import_module_all(module_name, namespace)
            else:
                module = importlib.import_module(module_name)
                namespace[attribute_name] = getattr(module, attribute_name)
        elif import_statement.startswith("import "):
            module_name = import_statement.split()[1]
            namespace[module_name] = importlib.import_module(module_name)
        else:
            raise ValueError("Instruction d'import invalide.")
        
        return namespace
    
    @staticmethod
    def import_module_all(module_name, namespace):
        module = importlib.import_module(module_name)

        for name in dir(module):
            if not name.startswith('_'):
                namespace[name] = getattr(module, name)
                
        return namespace
    
    @staticmethod
    def create_name_space(sender) -> dict:
        namespace = {}
        namespace['self'] = sender
        calling_module = inspect.getmodule(type(sender))
        if calling_module is None:
            return namespace
        source_code = inspect.getsource(calling_module)
        imports = re.findall(r'^(import .+|from .+ import .+)', source_code, re.MULTILINE)
        for imp in imports:
            namespace = PyExplorer.import_from_string(imp, namespace)
        return namespace